import os
import shutil
import stat
import subprocess
import tempfile

import pytest

from mentat.config_manager import ConfigManager
from mentat.streaming_printer import StreamingPrinter
from mentat.user_input_manager import UserInputManager

pytest_plugins = ("pytest_reportlog",)


def filter_mark(items, mark, exists):
    new_items = []
    for item in items:
        marker = item.get_closest_marker(mark)
        if bool(marker) == bool(exists):
            new_items.append(item)
    return new_items


def pytest_addoption(parser):
    parser.addoption("--benchmark", action="store_true")
    parser.addoption("--uitest", action="store_true")


def pytest_configure(config):
    config.addinivalue_line("markers", "benchmark: run benchmarks that call openai")
    config.addinivalue_line(
        "markers", "uitest: run ui-tests that get evaluated by humans"
    )


def pytest_collection_modifyitems(config, items):
    benchmark = config.getoption("--benchmark")
    uitest = config.getoption("--uitest")
    items[:] = filter_mark(items, "benchmark", benchmark)
    items[:] = filter_mark(items, "uitest", uitest)


@pytest.fixture
def mock_call_llm_api(mocker):
    mock = mocker.patch("mentat.parsing.call_llm_api")

    def set_generator_values(values):
        async def async_generator():
            for value in values:
                yield {"choices": [{"delta": {"content": value}}]}

        mock.return_value = async_generator()

    mock.set_generator_values = set_generator_values
    return mock


@pytest.fixture
def mock_collect_user_input(mocker):
    mock_method = mocker.MagicMock()

    mocker.patch.object(UserInputManager, "collect_user_input", new=mock_method)

    return mock_method


@pytest.fixture
def mock_setup_api_key(mocker):
    mocker.patch("mentat.app.setup_api_key")
    mocker.patch("mentat.conversation.check_model_availability")
    return


@pytest.fixture
def mock_config(temp_testbed):
    config = ConfigManager(temp_testbed)
    config.project_config = {}
    return config


def add_permissions(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.
    """

    # Is the error an access error?
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


# Auto-used fixtures


@pytest.fixture(autouse=True)
def temp_testbed(monkeypatch):
    # create temporary copy of testbed, complete with git repo
    # realpath() resolves symlinks, required for paths to match on macOS
    temp_dir = os.path.realpath(tempfile.mkdtemp())
    temp_testbed = os.path.join(temp_dir, "testbed")
    shutil.copytree("testbed", temp_testbed)
    shutil.copy(".gitignore", temp_testbed)
    subprocess.run(
        ["git", "init"],
        cwd=temp_testbed,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    subprocess.run(
        ["git", "add", "."],
        cwd=temp_testbed,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    subprocess.run(
        ["git", "commit", "-m", "add testbed"],
        cwd=temp_testbed,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    # necessary to undo chdir before calling rmtree, or it fails on windows
    with monkeypatch.context() as m:
        m.chdir(temp_testbed)
        yield temp_testbed

    shutil.rmtree(temp_dir, onerror=add_permissions)


@pytest.fixture(autouse=True)
def mock_sleep_time(mocker):
    mocker.patch.object(StreamingPrinter, "sleep_time", new=lambda self: 0)


# Creating a prompt session in Github Actions on Windows throws an error
# even though we don't use it, so we always have to mock the prompt session.
@pytest.fixture(autouse=True)
def mock_prompt_session(mocker):
    mocker.patch("mentat.user_input_manager.PromptSession")



from termcolor import cprint

from .code_change import CodeChange
from .code_file_manager import CodeFileManager
from .config_manager import ConfigManager
from .llm_api import CostTracker, check_model_availability, choose_model, count_tokens
from .parsing import run_async_stream_and_parse_llm_response
from .prompts import system_prompt


class Conversation:
    def __init__(
        self,
        config: ConfigManager,
        cost_tracker: CostTracker,
        code_file_manager: CodeFileManager,
    ):
        self.messages = []
        self.add_system_message(system_prompt)
        self.cost_tracker = cost_tracker
        self.code_file_manager = code_file_manager
        self.allow_32k = check_model_availability(config.allow_32k())

        tokens = count_tokens(code_file_manager.get_code_message()) + count_tokens(
            system_prompt
        )
        token_limit = 32768 if self.allow_32k else 8192
        if tokens > token_limit:
            raise KeyboardInterrupt(
                f"Included files already exceed token limit ({tokens} / {token_limit})."
                " Please try running again with a reduced number of files."
            )
        elif tokens + 1000 > token_limit:
            cprint(
                f"Warning: Included files are close to token limit ({tokens} /"
                f" {token_limit}), you may not be able to have a long conversation.",
                "red",
            )
        else:
            cprint(f"File and prompt token count: {tokens} / {token_limit}", "cyan")

    def add_system_message(self, message: str):
        self.messages.append({"role": "system", "content": message})

    def add_user_message(self, message: str):
        self.messages.append({"role": "user", "content": message})

    def add_assistant_message(self, message: str):
        self.messages.append({"role": "assistant", "content": message})

    def get_model_response(self, config: ConfigManager) -> (str, list[CodeChange]):
        messages = self.messages.copy()

        code_message = self.code_file_manager.get_code_message()
        messages.append({"role": "system", "content": code_message})

        model, num_prompt_tokens = choose_model(messages, self.allow_32k)

        state = run_async_stream_and_parse_llm_response(
            messages, model, self.code_file_manager
        )

        self.cost_tracker.display_api_call_stats(
            num_prompt_tokens,
            count_tokens(state.message),
            model,
            state.time_elapsed,
        )

        self.add_assistant_message(state.message)
        return state.explanation, state.code_changes



# Raised when model output doesn't adhere to the specified format for changes
class ModelError(Exception):
    def __init__(self, message, already_added_to_changelist):
        super().__init__(message)
        self.already_added_to_changelist = already_added_to_changelist


# Used to indicate an issue with Mentat's code
class MentatError(Exception):
    pass


# Used to indicate an issue with the user's usage of Mentat
class UserError(Exception):
    pass



import os
import shutil
import subprocess

import pytest

from mentat.app import run

# These benchmarks use GPT and won't run automatically.
# Run them with python tests/record_benchmark.py True
# to record the results or use pytest --benchmark
# to not save the results.
pytestmark = pytest.mark.benchmark


def test_calculator_add_power(mock_collect_user_input):
    mock_collect_user_input.side_effect = [
        (
            "Add power as a possible operation, raising the first arg to the power of"
            " the second"
        ),
        "y",
        KeyboardInterrupt,
    ]

    calculator_path = os.path.join("scripts/calculator.py")
    run([calculator_path])

    result = subprocess.run(
        ["python", calculator_path, "power", "15", "3"], capture_output=True, text=True
    )
    assert float(result.stdout.strip()) == 3375.0


def test_calculator_add_exp_then_log(mock_collect_user_input):
    mock_collect_user_input.side_effect = [
        "Add exponentation operation, called with 'exp'",
        "y",
        "Add logarithm operation, called with 'log'",
        "y",
        KeyboardInterrupt,
    ]

    calculator_path = "scripts/calculator.py"
    run([calculator_path])

    result = subprocess.run(
        ["python", calculator_path, "exp", "15", "3"], capture_output=True, text=True
    )
    assert float(result.stdout.strip()) == 3375.0

    result = subprocess.run(
        ["python", calculator_path, "log", "10", "2"], capture_output=True, text=True
    )
    assert float(result.stdout.strip()) == 3.3219280948873626


def test_calculator_continue_change(mock_collect_user_input):
    mock_collect_user_input.side_effect = [
        "complete the change I started",
        "y",
        KeyboardInterrupt,
    ]

    calculator_path = "scripts/calculator.py"

    with open(calculator_path, "r") as f:
        calculator_lines = f.readlines()
    index = calculator_lines.index("    return a / b\n")
    new_lines = [
        "\n",
        "\n",
        "def exp_numbers(a, b):\n",
        "    return a**b\n",
    ]
    calculator_lines = (
        calculator_lines[: index + 1] + new_lines + calculator_lines[index + 1 :]
    )
    with open(calculator_path, "w") as f:
        f.writelines(calculator_lines)

    run([calculator_path])

    result = subprocess.run(
        ["python", calculator_path, "exp", "15", "3"], capture_output=True, text=True
    )
    assert float(result.stdout.strip()) == 3375.0


def test_multifile_calculator(mock_collect_user_input):
    mock_collect_user_input.side_effect = [
        "add exp and log functions to take a^b and log a base b",
        "y",
        KeyboardInterrupt,
    ]

    multifile_calculator_path = "multifile_calculator"
    calculator_path = os.path.join(multifile_calculator_path, "calculator.py")

    run([multifile_calculator_path])
    result = subprocess.run(
        ["python", calculator_path, "exp", "15", "3"], capture_output=True, text=True
    )
    assert float(result.stdout.strip()) == 3375.0

    result = subprocess.run(
        ["python", calculator_path, "log", "10", "2"], capture_output=True, text=True
    )
    assert float(result.stdout.strip()) == 3.3219280948873626


def test_start_project_from_scratch(mock_collect_user_input):
    # Clear the testbed so we can test that it works with empty directories
    for item in os.listdir("."):
        if os.path.isfile(item):
            os.remove(item)
        elif os.path.isdir(item):
            if item != ".git":
                shutil.rmtree(item)

    mock_collect_user_input.side_effect = [
        "make a file that does fizzbuzz, named fizzbuzz.py, going up to 10",
        "y",
        KeyboardInterrupt,
    ]
    run(["."])

    fizzbuzz_path = "fizzbuzz.py"
    assert os.path.exists(fizzbuzz_path)

    result = subprocess.run(["python", fizzbuzz_path], capture_output=True, text=True)
    expected_output = "1\n2\nFizz\n4\nBuzz\nFizz\n7\n8\nFizz\nBuzz\n"
    assert result.stdout.strip() == expected_output.strip()



import os
import subprocess

import pytest

from mentat.errors import UserError
from mentat.git_handler import get_shared_git_root_for_paths


def test_no_paths_given(temp_testbed):
    # Get temp_testbed as the git root when given no paths
    git_root = get_shared_git_root_for_paths([])
    assert git_root == temp_testbed


def test_paths_given(temp_testbed):
    # Get temp_testbed when given directory in temp_testbed
    git_root = get_shared_git_root_for_paths(["scripts"])
    assert git_root == temp_testbed


def test_two_git_roots_given():
    # Exits when given 2 paths with separate git roots
    with pytest.raises(UserError) as e_info:
        os.makedirs("git_testing_dir")
        subprocess.run(["git", "init"], cwd="git_testing_dir")

        _ = get_shared_git_root_for_paths(["./", "git_testing_dir"])
    assert e_info.type == UserError



import sys

import pytest

from mentat.app import run
from mentat.user_input_manager import UserInputManager

# These ui-tests won't run automatically
# To run them, use pytest -s --uitest
# Without the -s, pytest will capture all
# i/o and the tests will not work
pytestmark = pytest.mark.uitest


def is_test_correct():
    ans = input("Does this look correct? (Y/n)").lower()
    sys.exit(0 if ans in ["", "y"] else 1)


@pytest.fixture
def ui_mock_collect_user_input(mocker):
    mock_method = mocker.MagicMock(side_effect=is_test_correct)
    mocker.patch.object(UserInputManager, "collect_user_input", new=mock_method)
    return mock_method


def test_start(mock_call_llm_api, ui_mock_collect_user_input):
    print()
    with pytest.raises(SystemExit) as e_info:
        run(["./"])
    assert e_info.value.code == 0



system_prompt = (
    "You are part of an automated coding system. As such, responses must adhere"
    " strictly to the required format, so they can be parsed programmaticaly. Your"
    " input will consist of a user request, the contents of code files, and sometimes"
    " the git diff of current code files."
    " The request may be to add a new feature, update the code, fix a"
    " bug, add comments or docstrings, etc.\nThe first part of your response should"
    " contain an brief summary of the changes you plan to make, then a list of the"
    " changes. Ensure you plan ahead, like planning to add imports for things you need"
    " to use in your changes, etc. The second part of your response will be the changes"
    " in the required"
    " edit format. Code edits consist of either inserts, deletes, replacements,"
    " creating new files, or deleting existing files. They"
    " can be of multiple lines of code. Edit description blocks start with @@start and"
    " end with @@end. If the edit is a delete or delete-file, then the block should"
    " only"
    " contain a"
    " JSON formatted section. In insert, replace, and create-file blocks, there must"
    " be a"
    " second"
    " section containing the new line or lines of code. The JSON section and code"
    " section are separated by a line containing just @@code.\nIf the request requires"
    " clarification or the user is asking for something other than code changes, such"
    " as design ideas, don't return any edit description blocks."
    """

To demonstrate the response format, here's an example user request, followed by an example response:


Code Files:

core/script.py
1:
2:def say_hello(name):
3:    print(f"Hello {name}!")
4:
5:
6:def say_goodbye():
7:    print("Goodbye!")
8:
9:
10:def main(name):
11:    say_hello(name)
12:    say_goodbye()
13:    print("Done!")
14:

core/hello_world.py
1:
2:def hello_world():
3:    print("Hello, World!")
4:

User Request:
After saying hello, if the user's name is "Bob", say "Nice to see you again!" on another line.
Add a function to get the user's name and use it in main instead of taking name as an argument.
The new function should be in a separate file called utils.py. Stop saying "Done!". Finally,
delete the hello_world.py file.


Example Response:

I will make the modifications to script.py and create the new file, importing from it in script.py.

Steps:
1. Modify say_hello, adding the case for Bob.
2. Create utils.py with a function to get the user's name.
3. Import the new function in script.py.
4. Modify main to use the new function instead of taking name as an argument.
5. Remove the line printing "Done!".
6. Delete file hello_world.py

@@start
{
    "file": "core/script.py",
    "action": "insert",
    "insert-after-line": 3,
    "insert-before-line": 4
}
@@code
    if name == "Bob":
        print("Nice to see you again!")
@@end
@@start
{
    "file": "core/utils.py",
    "action": "create-file"
}
@@code
def get_name():
    return input("Enter your name: ")
@@end
@@start
{
    "file": "core/script.py",
    "action": "insert",
    "insert-after-line": 0,
    "insert-before-line": 1
}
@@code
from core.utils import get_name
@@end
@@start
{
    "file": "core/script.py",
    "action": "replace",
    "start-line": 10,
    "end-line": 10
}
@@code
def main():
    name = get_name()
@@end
@@start
{
    "file": "core/script.py",
    "action": "delete",
    "start-line": 13,
    "end-line": 13
}
@@end
@@start
{
    "file": "core/hello_world.py",
    "action": "delete-file",
}
@@end
"""
)



import json
import logging
import os
from importlib import resources
from json import JSONDecodeError
from pathlib import Path

from termcolor import cprint

mentat_dir_path = os.path.join(Path.home(), ".mentat")

# package_name should always be "mentat" - but this will work if package name is changed
package_name = __name__.split(".")[0]
default_config_file_name = "default_config.json"
config_file_name = ".mentat_config.json"
user_config_path = os.path.join(mentat_dir_path, config_file_name)

# Remove this warning after August 19
old_config_file_path = os.path.join(mentat_dir_path, "config.json")


class ConfigManager:
    def __init__(self, git_root: str):
        # Remove this warning after August 19
        if os.path.exists(old_config_file_path):
            cprint(
                "Warning: You are still using an old config.json in your ~/.mentat"
                " directory. The config filename has recently been changed to"
                " .mentat_config.json, and can be present in either ~/.mentat or the"
                " git project you are working in. Your current config.json file will"
                " not be used.",
                color="light_yellow",
            )

        if os.path.exists(user_config_path):
            with open(user_config_path) as config_file:
                try:
                    self.user_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("User config file contains invalid json")
                    cprint(
                        "Warning: User .mentat_config.json contains invalid"
                        " json; ignoring user configuration file",
                        "light_yellow",
                    )
                    self.user_config = {}
        else:
            self.user_config = {}

        project_config_path = os.path.join(git_root, config_file_name)
        if os.path.exists(project_config_path):
            with open(project_config_path) as config_file:
                try:
                    self.project_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("Project config file contains invalid json")
                    cprint(
                        "Warning: Git project .mentat_config.json contains invalid"
                        " json; ignoring project configuration file",
                        "light_yellow",
                    )
                    self.project_config = {}
        else:
            self.project_config = {}

        default_config_path = resources.files(package_name).joinpath(
            default_config_file_name
        )
        with default_config_path.open("r") as config_file:
            self.default_config = json.load(config_file)

    def input_style(self) -> list[list[str]]:
        return self._get_key("input-style")

    def allow_32k(self) -> bool:
        return self._get_key("allow-32k")

    def file_exclude_glob_list(self) -> list[str]:
        return self._get_key("file-exclude-glob-list")

    def _get_key(self, key: str):
        if key in self.project_config:
            return self.project_config[key]
        elif key in self.user_config:
            return self.user_config[key]
        elif key in self.default_config:
            return self.default_config[key]
        else:
            raise ValueError(f"No value for config key {key} found")



import datetime
import logging
import logging.handlers
import os

from .config_manager import mentat_dir_path


def setup_logging():
    logs_dir = "logs"
    if "PYTEST_CURRENT_TEST" in os.environ:
        logs_dir += "/test_logs"
    logs_path = os.path.join(mentat_dir_path, logs_dir)

    logging.getLogger("openai").setLevel(logging.WARNING)
    # Breaking out of async generator when model messes up causes an error
    logging.getLogger("asyncio").setLevel(logging.CRITICAL)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    # Only log warnings and higher to console
    console_handler.setLevel(logging.WARNING)

    os.makedirs(logs_path, exist_ok=True)

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    log_file = os.path.join(logs_path, f"mentat_{timestamp}.log")
    latest_log_file = os.path.join(logs_path, "latest.log")
    if os.path.exists(latest_log_file):
        os.remove(latest_log_file)

    file_handler = logging.FileHandler(log_file)
    file_handler_latest = logging.FileHandler(latest_log_file)
    file_handler.setFormatter(formatter)
    file_handler_latest.setFormatter(formatter)

    costs_logger = logging.getLogger("costs")
    costs_formatter = logging.Formatter("%(asctime)s\n%(message)s")
    costs_handler = logging.FileHandler(os.path.join(logs_path, "costs.log"))
    costs_handler.setFormatter(costs_formatter)
    costs_logger.addHandler(costs_handler)
    costs_logger.setLevel(logging.INFO)
    costs_logger.propagate = False

    handlers = [console_handler, file_handler, file_handler_latest]
    logging.basicConfig(
        level=logging.DEBUG,
        handlers=handlers,
    )



import argparse
import glob
import logging
import os
from typing import Iterable, Optional

from termcolor import cprint

from .code_change import CodeChange
from .code_change_display import print_change
from .code_file_manager import CodeFileManager
from .config_manager import ConfigManager, mentat_dir_path
from .conversation import Conversation
from .errors import MentatError, UserError
from .git_handler import get_shared_git_root_for_paths
from .llm_api import CostTracker, setup_api_key
from .logging_config import setup_logging
from .user_input_manager import UserInputManager, UserQuitInterrupt


def run_cli():
    parser = argparse.ArgumentParser(
        description="Run conversation with command line args"
    )
    parser.add_argument(
        "paths",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns",
    )
    parser.add_argument(
        "--exclude",
        "-e",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns to exclude",
    )
    args = parser.parse_args()
    paths = args.paths
    exclude_paths = args.exclude
    run(expand_paths(paths), expand_paths(exclude_paths))


def expand_paths(paths: Iterable[str]) -> Iterable[str]:
    globbed_paths = set()
    invalid_paths = []
    for path in paths:
        new_paths = glob.glob(pathname=path, recursive=True)
        if new_paths:
            globbed_paths.update(new_paths)
        else:
            invalid_paths.append(path)
    if invalid_paths:
        cprint(
            "The following paths do not exist:",
            "light_yellow",
        )
        print("\n".join(invalid_paths))
        exit()
    return globbed_paths


def run(paths: Iterable[str], exclude_paths: Optional[Iterable[str]] = None):
    os.makedirs(mentat_dir_path, exist_ok=True)
    setup_logging()
    logging.debug(f"Paths: {paths}")

    cost_tracker = CostTracker()
    try:
        setup_api_key()
        loop(paths, exclude_paths, cost_tracker)
    except (
        EOFError,
        KeyboardInterrupt,
        UserQuitInterrupt,
        UserError,
        MentatError,
    ) as e:
        if str(e):
            cprint("\n" + str(e), "red")
    finally:
        cost_tracker.display_total_cost()


def loop(
    paths: Iterable[str],
    exclude_paths: Optional[Iterable[str]],
    cost_tracker: CostTracker,
) -> None:
    git_root = get_shared_git_root_for_paths(paths)
    config = ConfigManager(git_root)
    user_input_manager = UserInputManager(config)
    code_file_manager = CodeFileManager(
        paths,
        exclude_paths if exclude_paths is not None else [],
        user_input_manager,
        config,
        git_root,
    )
    conv = Conversation(config, cost_tracker, code_file_manager)

    cprint("Type 'q' or use Ctrl-C to quit at any time.\n", color="cyan")
    cprint("What can I do for you?", color="light_blue")
    need_user_request = True
    while True:
        if need_user_request:
            user_response = user_input_manager.collect_user_input()
            conv.add_user_message(user_response)
        explanation, code_changes = conv.get_model_response(config)

        if code_changes:
            need_user_request = get_user_feedback_on_changes(
                config, conv, user_input_manager, code_file_manager, code_changes
            )
        else:
            need_user_request = True


def get_user_feedback_on_changes(
    config: ConfigManager,
    conv: Conversation,
    user_input_manager: UserInputManager,
    code_file_manager: CodeFileManager,
    code_changes: Iterable[CodeChange],
) -> bool:
    cprint(
        "Apply these changes? 'Y/n/i' or provide feedback.",
        color="light_blue",
    )
    user_response = user_input_manager.collect_user_input()

    need_user_request = True
    match user_response.lower():
        case "y" | "":
            code_changes_to_apply = code_changes
            conv.add_user_message("User chose to apply all your changes.")
        case "n":
            code_changes_to_apply = []
            conv.add_user_message("User chose not to apply any of your changes.")
        case "i":
            code_changes_to_apply, indices = user_filter_changes(
                user_input_manager, code_changes
            )
            conv.add_user_message(
                "User chose to apply"
                f" {len(code_changes_to_apply)}/{len(code_changes)} of your suggest"
                " changes. The changes they applied were:"
                f" {', '.join(map(str, indices))}"
            )
        case _:
            need_user_request = False
            code_changes_to_apply = []
            conv.add_user_message(
                "User chose not to apply any of your changes. User response:"
                f" {user_response}\n\nPlease adjust your previous plan and changes to"
                " reflect this. Respond with a full new set of changes."
            )

    if code_changes_to_apply:
        code_file_manager.write_changes_to_files(code_changes_to_apply)
        if len(code_changes_to_apply) == len(code_changes):
            cprint("Changes applied.", color="light_blue")
        else:
            cprint("Selected changes applied.", color="light_blue")
    else:
        cprint("No changes applied.", color="light_blue")

    if need_user_request:
        cprint("Can I do anything else for you?", color="light_blue")

    return need_user_request


def user_filter_changes(
    user_input_manager: UserInputManager, code_changes: Iterable[CodeChange]
) -> Iterable[CodeChange]:
    new_changes = []
    indices = []
    for index, change in enumerate(code_changes, start=1):
        print_change(change)
        cprint("Keep this change?", "light_blue")
        if user_input_manager.ask_yes_no(default_yes=True):
            new_changes.append(change)
            indices.append(index)
    return new_changes, indices



import os
from enum import Enum
from pathlib import Path

from pygments.lexers import TextLexer, get_lexer_for_filename
from pygments.util import ClassNotFound

from .errors import ModelError


class CodeChangeAction(Enum):
    Insert = "insert"
    Replace = "replace"
    Delete = "delete"
    CreateFile = "create-file"
    DeleteFile = "delete-file"

    def has_surrounding_lines(self):
        return (
            self != CodeChangeAction.CreateFile and self != CodeChangeAction.DeleteFile
        )

    def has_removals(self):
        return (
            self == CodeChangeAction.Delete
            or self == CodeChangeAction.Replace
            or self == CodeChangeAction.DeleteFile
        )

    def has_additions(self):
        return (
            self == CodeChangeAction.Insert
            or self == CodeChangeAction.Replace
            or self == CodeChangeAction.CreateFile
        )


class CodeChange:
    def __init__(
        self,
        json_data: dict,
        code_lines: list[str],
        git_root: str,
        code_file_manager,
    ):
        self.json_data = json_data
        # Sometimes GPT puts quotes around numbers, so we have to convert those
        for json_key in [
            "insert-before-line",
            "insert-after-line",
            "start-line",
            "end-line",
        ]:
            if json_key in self.json_data:
                self.json_data[json_key] = int(self.json_data[json_key])
        self.code_lines = code_lines
        self.file = Path(self.json_data["file"])
        self.first_changed_line = None
        self.last_changed_line = None
        self.git_root = git_root
        self.error = False
        try:
            self.lexer = get_lexer_for_filename(self.file)
            self.lexer.stripnl = False
            self.lexer.stripall = False
            self.lexer.ensurenl = False
        except ClassNotFound:
            self.lexer = TextLexer()

        try:
            self.action = CodeChangeAction(self.json_data["action"])
        except ValueError:
            raise ModelError(
                f"Model created change with unknown action {self.json_data['action']}",
                already_added_to_changelist=False,
            )

        try:
            match self.action:
                case CodeChangeAction.Insert:
                    if "insert-before-line" in self.json_data:
                        self.first_changed_line = self.json_data["insert-before-line"]
                        if "insert-after-line" in self.json_data:
                            if (
                                self.first_changed_line - 1
                                != self.json_data["insert-after-line"]
                            ):
                                self.error = "Insert line numbers invalid"
                    elif "insert-after-line" in self.json_data:
                        self.first_changed_line = (
                            self.json_data["insert-after-line"] + 1
                        )
                    else:
                        self.error = "Insert line number not specified"
                    self.first_changed_line -= 0.5
                    self.last_changed_line = self.first_changed_line

                case CodeChangeAction.Replace:
                    self.first_changed_line = self.json_data["start-line"]
                    self.last_changed_line = self.json_data["end-line"]

                case CodeChangeAction.Delete:
                    self.first_changed_line = self.json_data["start-line"]
                    self.last_changed_line = self.json_data["end-line"]
        except KeyError:
            self.error = "Line numbers not given"

        if (
            self.first_changed_line
            and self.last_changed_line
            and self.first_changed_line > self.last_changed_line
        ):
            self.error = "Starting line of change is greater than ending line of change"

        if self.action != CodeChangeAction.CreateFile:
            file_path = os.path.join(self.git_root, self.file)
            try:
                self.file_lines = code_file_manager.file_lines[file_path]
                self.line_number_buffer = len(str(len(self.file_lines) + 1)) + 1
            except KeyError:
                self.error = (
                    f"Model attempted to edit {file_path}, which isn't in current"
                    " context or doesn't exist"
                )
        else:
            if os.path.exists(self.file):
                self.error = (
                    f"Model attempted to create file that already exists: {self.file}"
                )

            self.file_lines = []
            self.line_number_buffer = 2

    def __lt__(self, other):
        return self.last_changed_line < other.last_changed_line

    def apply(self, cur_file_lines: list[str]) -> list[str]:
        match self.action:
            case CodeChangeAction.Insert:
                previous_lines = cur_file_lines[: int(self.first_changed_line)]
                following_lines = cur_file_lines[int(self.first_changed_line) :]
                new_file_lines = previous_lines + self.code_lines + following_lines

            case CodeChangeAction.Replace:
                previous_lines = cur_file_lines[: self.first_changed_line - 1]
                following_lines = cur_file_lines[self.last_changed_line :]
                new_file_lines = previous_lines + self.code_lines + following_lines

            case CodeChangeAction.Delete:
                previous_lines = cur_file_lines[: self.first_changed_line - 1]
                following_lines = cur_file_lines[self.last_changed_line :]
                new_file_lines = previous_lines + following_lines

            case CodeChangeAction.CreateFile | CodeChangeAction.DeleteFile:
                raise Exception(
                    f"CodeChange with action={self.action} shouldn't have apply called"
                )

        return new_file_lines



import logging
import os
import sys
from dataclasses import dataclass
from typing import Generator

import openai
import tiktoken
from dotenv import load_dotenv
from termcolor import cprint

from .config_manager import mentat_dir_path, user_config_path
from .errors import MentatError, UserError

package_name = __name__.split(".")[0]


# Check for .env file or already exported API key
# If no api key found, raise an error
def setup_api_key():
    if not load_dotenv(os.path.join(mentat_dir_path, ".env")):
        load_dotenv()
    key = os.getenv("OPENAI_API_KEY")
    if not key:
        raise UserError(
            "No OpenAI api key detected.\nEither place your key into a .env"
            " file or export it as an environment variable."
        )
    try:
        openai.api_key = key
        openai.Model.list()  # Test the API key
    except openai.error.AuthenticationError as e:
        raise UserError(f"OpenAI gave an Authentication Error:\n{e}")


async def call_llm_api(messages: list[dict[str, str]], model) -> Generator:
    if (
        "PYTEST_CURRENT_TEST" in os.environ
        and "--benchmark" not in sys.argv
        and os.getenv("MENTAT_BENCHMARKS_RUNNING") == "false"
    ):
        logging.critical("OpenAI call attempted in non benchmark test environment!")
        raise MentatError("OpenAI call attempted in non benchmark test environment!")

    response = await openai.ChatCompletion.acreate(
        model=model,
        messages=messages,
        temperature=0.5,
        stream=True,
    )

    return response


def count_tokens(message: str) -> int:
    return len(
        tiktoken.encoding_for_model("gpt-4").encode(message, disallowed_special=())
    )


def check_model_availability(allow_32k: bool) -> bool:
    available_models = [x["id"] for x in openai.Model.list()["data"]]
    if allow_32k:
        # check if user has access to gpt-4-32k
        if "gpt-4-32k-0314" not in available_models:
            cprint(
                "You set ALLOW_32K to true, but your OpenAI API key doesn't"
                " have access to gpt-4-32k-0314. To remove this warning, set"
                " ALLOW_32K to false until you have access.",
                "yellow",
            )
            allow_32k = False

    if not allow_32k:
        # check if user has access to gpt-4
        if "gpt-4-0314" not in available_models:
            raise UserError(
                "Sorry, but your OpenAI API key doesn't have access to gpt-4-0314,"
                " which is currently required to run Mentat."
            )

    return allow_32k


def choose_model(messages: list[dict[str, str]], allow_32k) -> str:
    prompt_token_count = 0
    for message in messages:
        prompt_token_count += count_tokens(message["content"])
    cprint(f"\nTotal token count: {prompt_token_count}", "cyan")

    model = "gpt-4-0314"
    token_buffer = 500
    if prompt_token_count > 8192 - token_buffer:
        if allow_32k:
            model = "gpt-4-32k-0314"
            if prompt_token_count > 32768 - token_buffer:
                cprint(
                    "Warning: gpt-4-32k-0314 has a token limit of 32768. Attempting"
                    " to run anyway:"
                )
        else:
            cprint(
                "Warning: gpt-4-0314 has a maximum context length of 8192 tokens."
                " If you have access to gpt-4-32k-0314, set allow-32k to `true` in"
                f" `{user_config_path}` to use"
                " it. Attempting to run with gpt-4-0314:",
                "yellow",
            )
    return model, prompt_token_count


@dataclass
class CostTracker:
    total_cost: int = 0

    def display_api_call_stats(
        self,
        num_prompt_tokens: int,
        num_sampled_tokens: int,
        model: str,
        call_time: float,
    ) -> None:
        cost_per_1000_tokens = {
            "gpt-4-0314": (0.03, 0.06),
            "gpt-4-32k-0314": (0.06, 0.12),
        }
        prompt_cost = (num_prompt_tokens / 1000) * cost_per_1000_tokens[model][0]
        sampled_cost = (num_sampled_tokens / 1000) * cost_per_1000_tokens[model][1]

        tokens_per_second = num_sampled_tokens / call_time
        call_cost = prompt_cost + sampled_cost

        speed_and_cost_string = (
            f"Speed: {tokens_per_second:.2f} tkns/s | Cost: ${call_cost:.2f}"
        )
        cprint(speed_and_cost_string, "cyan")

        costs_logger = logging.getLogger("costs")
        costs_logger.info(speed_and_cost_string)

        self.total_cost += call_cost

    def display_total_cost(self) -> None:
        cprint(f"\nTotal session cost: ${self.total_cost:.2f}", color="light_blue")



import glob
import logging
import math
import os
from collections import defaultdict
from pathlib import Path
from typing import Iterable

from termcolor import cprint

from .change_conflict_resolution import (
    resolve_insertion_conflicts,
    resolve_non_insertion_conflicts,
)
from .code_change import CodeChange, CodeChangeAction
from .config_manager import ConfigManager
from .errors import MentatError, UserError
from .git_handler import (
    get_git_diff_for_path,
    get_non_gitignored_files,
    get_paths_with_git_diffs,
)
from .user_input_manager import UserInputManager


def _build_path_tree(file_paths, git_root):
    tree = {}
    for path in file_paths:
        path = os.path.relpath(path, git_root)
        parts = Path(path).parts
        current_level = tree
        for part in parts:
            if part not in current_level:
                current_level[part] = {}
            current_level = current_level[part]
    return tree


def _print_path_tree(tree, changed_files, cur_path, prefix=""):
    keys = list(tree.keys())
    for i, key in enumerate(sorted(keys)):
        if i < len(keys) - 1:
            new_prefix = prefix + "│   "
            print(f"{prefix}├── ", end="")
        else:
            new_prefix = prefix + "    "
            print(f"{prefix}└── ", end="")

        cur = os.path.join(cur_path, key)
        star = "* " if cur in changed_files else ""
        if tree[key]:
            color = "blue"
        elif star:
            color = "green"
        else:
            color = None
        cprint(f"{star}{key}", color)
        if tree[key]:
            _print_path_tree(tree[key], changed_files, cur, new_prefix)


def _is_file_text_encoded(file_path):
    try:
        # The ultimate filetype test
        with open(file_path) as f:
            f.read()
        return True
    except UnicodeDecodeError:
        return False


def _abs_file_paths_from_list(paths: Iterable[str], check_for_text: bool = True):
    file_paths_direct = set()
    file_paths_from_dirs = set()
    for path in paths:
        path = Path(path)
        if path.is_file():
            if check_for_text and not _is_file_text_encoded(path):
                logging.info(f"File path {path} is not text encoded.")
                raise UserError(f"File path {path} is not text encoded.")
            file_paths_direct.add(os.path.realpath(path))
        elif path.is_dir():
            nonignored_files = set(
                map(
                    lambda f: os.path.realpath(os.path.join(path, f)),
                    get_non_gitignored_files(path),
                )
            )

            file_paths_from_dirs.update(
                filter(
                    lambda f: (not check_for_text) or _is_file_text_encoded(f),
                    nonignored_files,
                )
            )
    return file_paths_direct, file_paths_from_dirs


class CodeFileManager:
    def __init__(
        self,
        paths: Iterable[str],
        exclude_paths: Iterable[str],
        user_input_manager: UserInputManager,
        config: ConfigManager,
        git_root: str,
    ):
        self.config = config
        self.git_root = git_root
        self._set_file_paths(paths, exclude_paths)
        self.user_input_manager = user_input_manager

        if self.file_paths:
            cprint("Files included in context:", "green")
        else:
            cprint("No files included in context.\n", "red")
            cprint("Git project: ", "green", end="")
        cprint(os.path.split(self.git_root)[1], "blue")
        _print_path_tree(
            _build_path_tree(self.file_paths, self.git_root),
            get_paths_with_git_diffs(self.git_root),
            self.git_root,
        )
        print()

    def _set_file_paths(
        self, paths: Iterable[str], exclude_paths: Iterable[str]
    ) -> None:
        excluded_files, excluded_files_from_dir = _abs_file_paths_from_list(
            exclude_paths, check_for_text=False
        )

        glob_excluded_files = set(
            os.path.join(self.git_root, file)
            for glob_path in self.config.file_exclude_glob_list()
            # If the user puts a / at the beginning, it will try to look in root directory
            for file in glob.glob(
                pathname=glob_path,
                root_dir=self.git_root,
                recursive=True,
            )
        )
        file_paths_direct, file_paths_from_dirs = _abs_file_paths_from_list(
            paths, check_for_text=True
        )

        # config glob excluded files only apply to files added from directories
        file_paths_from_dirs -= glob_excluded_files

        self.file_paths = list(
            (file_paths_direct | file_paths_from_dirs)
            - (excluded_files | excluded_files_from_dir)
        )

    def _read_file(self, abs_path) -> Iterable[str]:
        with open(abs_path, "r") as f:
            lines = f.read().split("\n")
        return lines

    def _read_all_file_lines(self) -> None:
        self.file_lines = dict()
        for abs_path in self.file_paths:
            self.file_lines[abs_path] = self._read_file(abs_path)

    def get_code_message(self):
        self._read_all_file_lines()
        code_message = ["Code Files:\n"]
        for abs_path in self.file_paths:
            rel_path = os.path.relpath(abs_path, self.git_root)

            # We always want to give GPT posix paths
            posix_rel_path = Path(rel_path).as_posix()
            code_message.append(posix_rel_path)

            for i, line in enumerate(self.file_lines[abs_path], start=1):
                code_message.append(f"{i}:{line}")
            code_message.append("")

            git_diff_output = get_git_diff_for_path(self.git_root, rel_path)
            if git_diff_output:
                code_message.append("Current git diff for this file:")
                code_message.append(f"{git_diff_output}")

        return "\n".join(code_message)

    def _handle_delete(self, delete_change):
        file_path = os.path.join(self.git_root, delete_change.file)
        if not os.path.exists(file_path):
            logging.error(f"Path {file_path} non-existent on delete")
            return

        cprint(f"Are you sure you want to delete {delete_change.file}?", "red")
        if self.user_input_manager.ask_yes_no(default_yes=False):
            logging.info(f"Deleting file {file_path}")
            cprint(f"Deleting {delete_change.file}...")
            self.file_paths.remove(file_path)
            os.remove(file_path)
        else:
            cprint(f"Not deleting {delete_change.file}")

    def _get_new_code_lines(self, changes) -> Iterable[str]:
        if len(set(map(lambda change: change.file, changes))) > 1:
            raise Exception("All changes passed in must be for the same file")

        changes = sorted(changes, reverse=True)

        # We resolve insertion conflicts twice because non-insertion conflicts
        # might move insert blocks outside of replace/delete blocks and cause
        # them to conflict again
        changes = resolve_insertion_conflicts(changes, self.user_input_manager, self)
        changes = resolve_non_insertion_conflicts(changes, self.user_input_manager)
        changes = resolve_insertion_conflicts(changes, self.user_input_manager, self)
        if not changes:
            return []

        rel_path = changes[0].file
        abs_path = os.path.join(self.git_root, rel_path)
        new_code_lines = self.file_lines[abs_path].copy()
        if new_code_lines != self._read_file(abs_path):
            logging.info(f"File '{rel_path}' changed while generating changes")
            cprint(
                f"File '{rel_path}' changed while generating; current file changes"
                " will be erased. Continue?",
                color="light_yellow",
            )
            if not self.user_input_manager.ask_yes_no(default_yes=False):
                cprint(f"Not applying changes to file {rel_path}.")
                return None

        # Necessary in case the model needs to insert past the end of the file
        last_line = len(new_code_lines) + 1
        largest_changed_line = math.ceil(changes[0].last_changed_line)
        if largest_changed_line > last_line:
            new_code_lines += [""] * (largest_changed_line - last_line)

        min_changed_line = largest_changed_line + 1
        for i, change in enumerate(changes):
            if change.last_changed_line >= min_changed_line:
                raise MentatError(f"Change line number overlap in file {change.file}")
            min_changed_line = change.first_changed_line
            new_code_lines = change.apply(new_code_lines)
        return new_code_lines

    def write_changes_to_files(self, code_changes: list[CodeChange]) -> None:
        files_to_write = dict()
        file_changes = defaultdict(list)
        for code_change in code_changes:
            rel_path = code_change.file
            if code_change.action == CodeChangeAction.CreateFile:
                cprint(f"Creating new file {rel_path}", color="light_green")
                files_to_write[rel_path] = code_change.code_lines
            elif code_change.action == CodeChangeAction.DeleteFile:
                self._handle_delete(code_change)
            else:
                file_changes[rel_path].append(code_change)

        for file_path, changes in file_changes.items():
            new_code_lines = self._get_new_code_lines(changes)
            if new_code_lines:
                files_to_write[file_path] = new_code_lines

        for rel_path, code_lines in files_to_write.items():
            file_path = os.path.join(self.git_root, rel_path)
            if file_path not in self.file_paths:
                # newly created files added to Mentat's context
                logging.info(f"Adding new file {file_path} to context")
                self.file_paths.append(file_path)
                # create any missing directories in the path
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, "w") as f:
                f.write("\n".join(code_lines))



from textwrap import dedent

from mentat.app import run

temp_file_name = "temp.py"
template_insert_content = "# I inserted this comment"
template_insert_expected_content = template_insert_content + "\n"
template_double_insert_expected_content = (
    template_insert_content + "\n\n" + template_insert_content
)
template_insert = dedent(f"""\
    @@start
    {{
        "file": "{temp_file_name}",
        "action": "insert",
        "insert-after-line": 0,
        "insert-before-line": 1
    }}
    @@code
    {template_insert_content}
    @@end
    """)
template_insert2 = dedent(f"""\
    @@start
    {{
        "file": "{temp_file_name}",
        "action": "insert",
        "insert-after-line": 1,
        "insert-before-line": 2
    }}
    @@code
    {template_insert_content}
    @@end
    """)


def error_test_template(
    mock_call_llm_api,
    mock_collect_user_input,
    mock_setup_api_key,
    changes,
):
    # Automatically set everything up given and use given changes
    with open(temp_file_name, "w") as f:
        f.write("")

    mock_collect_user_input.side_effect = [
        "Go!",
        "y",
        KeyboardInterrupt,
    ]
    mock_call_llm_api.set_generator_values([changes])

    run([temp_file_name])
    with open(temp_file_name, "r") as f:
        content = f.read()
    return content


# These tests should not accept any changes after the invalid one
def test_malformed_json(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Should stop and only allow applying changes up to that point, not including malformed change
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent("""\
        @@start
        {{
            "malformed-json: [,
        }}
        @@code
        # My json is malformed :(
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content


def test_unknown_action(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Should stop and only allow applying changes up to that point, not including unknown action change
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "unknown",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        # I am unknown
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content


def test_indicator(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Test start indicator in change, code indicator outside change, code indicator inside code block,
    # code indicator for action without code, and end indicator while not creating change
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        @@start
        # I am wrong
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content

    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@code
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        # I am wrong
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content

    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        @@code
        # I am wrong
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content

    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "remove",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        # I am wrong
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content

    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent("""\
        @@end
        """) + template_insert2,
    )
    assert content == template_insert_expected_content


# These tests should keep changes after the invalid change
def test_no_line_numbers(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Should have line numbers
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "replace"
        }}
        @@code
        # I have no line numbers
        @@end
        """) + template_insert2,
    )
    assert content == template_double_insert_expected_content


def test_invalid_line_numbers(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # First line number should be <= the last line number
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "replace",
            "start-line": 10,
            "end-line": 4
        }}
        @@code
        # I have wrong line numbers
        @@end
        """) + template_insert2,
    )
    print(repr(content))
    assert content == template_double_insert_expected_content


def test_existing_file(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Creating file that already exists should fail
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent(f"""\
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "create-file"
        }}
        @@code
        # I already exist
        @@end
        """) + template_insert2,
    )
    print(repr(content))
    assert content == template_double_insert_expected_content


def test_file_not_in_context(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Trying to access file not in context should fail
    content = error_test_template(
        mock_call_llm_api,
        mock_collect_user_input,
        mock_setup_api_key,
        template_insert + dedent("""\
        @@start
        {
            "file": "iamnotincontext",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # I am not in context
        @@end
        """) + template_insert2,
    )
    print(repr(content))
    assert content == template_double_insert_expected_content



import logging
import os
import subprocess
from pathlib import Path

from mentat.errors import UserError


def get_git_diff_for_path(git_root, path: str) -> str:
    return subprocess.check_output(["git", "diff", path], cwd=git_root).decode("utf-8")


def get_non_gitignored_files(path: str) -> set[str]:
    return set(
        # git returns / separated paths even on windows, convert so we can remove
        # glob_excluded_files, which have windows paths on windows
        os.path.normpath(p)
        for p in filter(
            lambda p: p != "",
            subprocess.check_output(
                # -c shows cached (regular) files, -o shows other (untracked/new) files
                ["git", "ls-files", "-c", "-o", "--exclude-standard"],
                cwd=path,
                text=True,
            ).split("\n"),
        )
    )


def get_paths_with_git_diffs(git_root) -> set[str]:
    changed = subprocess.check_output(
        ["git", "diff", "--name-only"], cwd=git_root, text=True
    ).split("\n")
    new = subprocess.check_output(
        ["git", "ls-files", "-o", "--exclude-standard"], cwd=git_root, text=True
    ).split("\n")
    return set(
        map(
            lambda path: os.path.realpath(os.path.join(git_root, Path(path))),
            changed + new,
        )
    )


def _get_git_root_for_path(path) -> str:
    if os.path.isdir(path):
        dir_path = path
    else:
        dir_path = os.path.dirname(path)
    try:
        git_root = (
            subprocess.check_output(
                [
                    "git",
                    "rev-parse",
                    "--show-toplevel",
                ],
                cwd=os.path.realpath(dir_path),
                stderr=subprocess.DEVNULL,
            )
            .decode("utf-8")
            .strip()
        )
        # call realpath to resolve symlinks, so all paths match
        return os.path.realpath(git_root)
    except subprocess.CalledProcessError:
        logging.error(f"File {path} isn't part of a git project.")
        raise UserError()


def get_shared_git_root_for_paths(paths) -> str:
    git_roots = set()
    for path in paths:
        git_root = _get_git_root_for_path(path)
        git_roots.add(git_root)
    if not paths:
        git_root = _get_git_root_for_path(os.getcwd())
        git_roots.add(git_root)

    if len(git_roots) > 1:
        logging.error(
            "All paths must be part of the same git project! Projects provided:"
            f" {git_roots}"
        )
        raise UserError()
    elif len(git_roots) == 0:
        logging.error("No git projects provided.")
        raise UserError()

    return git_roots.pop()



from textwrap import dedent

import pytest

from mentat.config_manager import ConfigManager, config_file_name


def test_config_priority(temp_testbed):
    # First project config should be considered, then user config, then default config, then error
    with open(config_file_name, "w") as project_config_file:
        project_config_file.write(dedent("""\
        {
            "project-first": "project"
        }"""))

    # Since we don't want to actually change user config file or default config file,
    # we have to just mock them
    config = ConfigManager(temp_testbed)
    config.user_config = {"project-first": "I will not be used", "user-second": "user"}
    config.default_config = {
        "project-first": "I also am not used",
        "user-second": "Neither am I",
        "default-last": "default",
    }

    assert config._get_key("project-first") == "project"
    assert config._get_key("user-second") == "user"
    assert config._get_key("default-last") == "default"

    with pytest.raises(ValueError) as e_info:
        config._get_key("non-existent")
    assert e_info.type == ValueError


def test_invalid_config(temp_testbed):
    # If invalid config file is found, it should use next config
    with open(config_file_name, "w") as project_config_file:
        project_config_file.write(dedent("""\
        {
            "mykey": "project",
            "invalid-json: []]
        }"""))

    config = ConfigManager(temp_testbed)
    config.user_config = {"mykey": "user"}
    assert config._get_key("mykey") == "user"
